# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Purpose
    Demonstrate basic bucket operations in Amazon S3.
    Learn how to create, list, delete, and configure buckets.
    Usage is shown in the test/test_bucket_wrapper.py file.

Running the tests
    The best way to learn how to use this service is to run the tests.
    For instructions on testing, see the README.

Running the code
    Run the usage_demo function in a command window or individual functions in
    the Python shell to make calls to your AWS account.
    For instructions on running the code, see the README.

Additional information
    Running this code might result in charges to your AWS account.
"""

import json
import logging
import uuid

import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)

s3_resource = boto3.resource('s3')


def get_s3(region=None):
    """Get a Boto 3 S3 resource with a specific Region or with your default Region."""
    global s3_resource
    if not region or s3_resource.meta.client.meta.region_name == region:
        return s3_resource
    else:
        return boto3.resource('s3', region_name=region)

def get_bucket(name,region=None):
    s3 = get_s3(region)
    try:
        bucket = s3.Bucket(name)
    except ClientError as error:
        logger.warning("Bucket %s doesn't exist or you don't have access to it.",
                       name)
        raise error
    return bucket

def create_bucket(name, region=None):
    """
    Create an Amazon S3 bucket with the specified name and in the specified Region.

    Usage is shown in usage_demo at the end of this module.

    :param name: The name of the bucket to create. This name must be globally unique
                 and must adhere to bucket naming requirements.
    :param region: The Region in which to create the bucket. If this is not specified,
                   the Region configured in your shared credentials is used. If no
                   Region is configured, 'us-east-1' is used.
    :return: The newly created bucket.
    """
    s3 = get_s3(region)

    try:
        if region:
            bucket = s3.create_bucket(
                Bucket=name,
                CreateBucketConfiguration={
                    'LocationConstraint': region
                }
            )
        else:
            bucket = s3.create_bucket(Bucket=name)

        bucket.wait_until_exists()

        logger.info("Created bucket '%s' in region=%s", bucket.name,
                    s3.meta.client.meta.region_name)
    except ClientError as error:
        logger.exception("Couldn't create bucket named '%s' in region=%s.",
                         name, region)
        if error.response['Error']['Code'] == 'IllegalLocationConstraintException':
            logger.error("When the session Region is anything other than us-east-1, "
                         "you must specify a LocationConstraint that matches the "
                         "session Region. The current session Region is %s and the "
                         "LocationConstraint Region is %s.",
                         s3.meta.client.meta.region_name, region)
        raise error
    else:
        return bucket


def bucket_exists(bucket_name):
    """
    Determine whether a bucket with the specified name exists.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to check.
    :return: True when the bucket exists; otherwise, False.
    """
    s3 = get_s3()
    try:
        s3.meta.client.head_bucket(Bucket=bucket_name)
        logger.info("Bucket %s exists.", bucket_name)
        exists = True
    except ClientError:
        logger.warning("Bucket %s doesn't exist or you don't have access to it.",
                       bucket_name)
        exists = False
    return exists


def get_buckets():
    """
    Get the buckets in all Regions for the current account.

    Usage is shown in usage_demo at the end of this module.

    :return: The list of buckets.
    """
    s3 = get_s3()
    try:
        buckets = list(s3.buckets.all())
        logger.info("Got buckets: %s.", buckets)
    except ClientError:
        logger.exception("Couldn't get buckets.")
        raise
    else:
        return buckets


def delete_bucket(bucket):
    """
    Delete a bucket. The bucket must be empty or an error is raised.

    Usage is shown in usage_demo at the end of this module.

    :param bucket: The bucket to delete.
    """
    try:
        bucket.delete()
        bucket.wait_until_not_exists()
        logger.info("Bucket %s successfully deleted.", bucket.name)
    except ClientError:
        logger.exception("Couldn't delete bucket %s.", bucket.name)
        raise


def grant_log_delivery_access(bucket_name):
    """
    Grant the AWS Log Delivery group write access to the specified bucket so that
    Amazon S3 can deliver access logs to the bucket. This is the only recommended
    use of an S3 bucket ACL.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to receive access logs.
    """

    s3 = get_s3()
    try:
        acl = s3.Bucket(bucket_name).Acl()
        # Putting an ACL overwrites the existing ACL. If you want to preserve
        # existing grants, append new grants to the list of existing grants.
        grants = acl.grants if acl.grants else []
        grants.append({
            'Grantee': {
                'Type': 'Group',
                'URI': 'http://acs.amazonaws.com/groups/s3/LogDelivery'
            },
            'Permission': 'WRITE'
        })
        acl.put(
            AccessControlPolicy={
                'Grants': grants,
                'Owner': acl.owner
            }
        )
        logger.info("Granted log delivery access to bucket '%s'", bucket_name)
    except ClientError:
        logger.exception("Couldn't add ACL to bucket '%s'.", bucket_name)
        raise


def get_acl(bucket_name):
    """
    Get the ACL of the specified bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to retrieve.
    :return: The ACL of the bucket.
    """
    s3 = get_s3()
    try:
        acl = s3.Bucket(bucket_name).Acl()
        logger.info("Got ACL for bucket %s owned by %s.",
                    bucket_name, acl.owner['DisplayName'])
    except ClientError:
        logger.exception("Couldn't get ACL for bucket %s.", bucket_name)
        raise
    else:
        return acl


def put_cors(bucket_name, cors_rules):
    """
    Apply CORS rules to a bucket. CORS rules specify the HTTP actions that are
    allowed from other domains.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket where the rules are applied.
    :param cors_rules: The CORS rules to apply.
    """
    s3 = get_s3()
    try:
        s3.Bucket(bucket_name).Cors().put(CORSConfiguration={
            'CORSRules': cors_rules
        })
        logger.info("Put CORS rules %s for bucket '%s'.", cors_rules, bucket_name)
    except ClientError:
        logger.exception("Couldn't put CORS rules for bucket %s.", bucket_name)
        raise


def get_cors(bucket_name):
    """
    Get the CORS rules for the specified bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to check.
    :return The CORS rules for the specified bucket.
    """
    s3 = get_s3()
    try:
        cors = s3.Bucket(bucket_name).Cors()
        logger.info("Got CORS rules %s for bucket '%s'.", cors.cors_rules, bucket_name)
    except ClientError:
        logger.exception(("Couldn't get CORS for bucket %s.", bucket_name))
        raise
    else:
        return cors


def delete_cors(bucket_name):
    """
    Delete the CORS rules from the specified bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to update.
    """
    s3 = get_s3()
    try:
        s3.Bucket(bucket_name).Cors().delete()
        logger.info("Deleted CORS from bucket '%s'.", bucket_name)
    except ClientError:
        logger.exception("Couldn't delete CORS from bucket '%s'.", bucket_name)
        raise


def put_policy(bucket_name, policy):
    """
    Apply a security policy to a bucket. Policies typically grant users the ability
    to perform specific actions, such as listing the objects in the bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to receive the policy.
    :param policy: The policy to apply to the bucket.
    """
    s3 = get_s3()
    try:
        # The policy must be in JSON format.
        s3.Bucket(bucket_name).Policy().put(Policy=json.dumps(policy))
        logger.info("Put policy %s for bucket '%s'.", policy, bucket_name)
    except ClientError:
        logger.exception("Couldn't apply policy to bucket '%s'.", bucket_name)
        raise


def get_policy(bucket_name):
    """
    Get the security policy of a bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The bucket to retrieve.
    :return: The security policy of the specified bucket.
    """
    s3 = get_s3()
    try:
        policy = s3.Bucket(bucket_name).Policy()
        logger.info("Got policy %s for bucket '%s'.", policy.policy, bucket_name)
    except ClientError:
        logger.exception("Couldn't get policy for bucket '%s'.", bucket_name)
        raise
    else:
        return json.loads(policy.policy)


def delete_policy(bucket_name):
    """
    Delete the security policy from the specified bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to update.
    """
    s3 = get_s3()
    try:
        s3.Bucket(bucket_name).Policy().delete()
        logger.info("Deleted policy for bucket '%s'.", bucket_name)
    except ClientError:
        logger.exception("Couldn't delete policy for bucket '%s'.", bucket_name)
        raise


def put_lifecycle_configuration(bucket_name, lifecycle_rules):
    """
    Apply a lifecycle configuration to a bucket. The lifecycle configuration can
    be used to archive or delete the objects in the bucket according to specified
    parameters, such as a number of days.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to update.
    :param lifecycle_rules: The lifecycle rules to apply.
    """
    s3 = get_s3()
    try:
        s3.Bucket(bucket_name).LifecycleConfiguration().put(
            LifecycleConfiguration={
                'Rules': lifecycle_rules
            }
        )
        logger.info("Put lifecycle rules %s for bucket '%s'.", lifecycle_rules,
                    bucket_name)
    except ClientError:
        logger.exception("Couldn't put lifecycle rules for bucket '%s'.", bucket_name)
        raise


def get_lifecycle_configuration(bucket_name):
    """
    Get the lifecycle configuration of the specified bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to retrieve.
    :return: The lifecycle rules of the specified bucket.
    """
    s3 = get_s3()
    try:
        config = s3.Bucket(bucket_name).LifecycleConfiguration()
        logger.info("Got lifecycle rules %s for bucket '%s'.",
                    config.rules, bucket_name)
    except:
        logger.exception("Couldn't get lifecycle configuration for bucket '%s'.",
                         bucket_name)
        raise
    else:
        return config.rules


def delete_lifecycle_configuration(bucket_name):
    """
    Remove the lifecycle configuration from the specified bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket to update.
    """
    s3 = get_s3()
    try:
        s3.Bucket(bucket_name).LifecycleConfiguration().delete()
        logger.info("Deleted lifecycle configuration for bucket '%s'.", bucket_name)
    except ClientError:
        logger.exception("Couldn't delete lifecycle configuration for bucket '%s'.",
                         bucket_name)
        raise


def generate_presigned_post(bucket_name, object_key, expires_in):
    """
    Generate a presigned Amazon S3 POST request to upload a file.
    A presigned POST can be used for a limited time to let someone without an AWS
    account upload a file to a bucket.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket that receives the posted object.
    :param object_key: The object key to identify the uploaded object.
    :param expires_in: The number of seconds the presigned POST is valid.
    :return: A dictionary that contains the URL and form fields that contain
             required access data.
    """
    s3 = get_s3()
    try:
        response = s3.meta.client.generate_presigned_post(
            Bucket=bucket_name, Key=object_key, ExpiresIn=expires_in)
        logger.info("Got presigned POST URL: %s", response['url'])
    except ClientError:
        logger.exception("Couldn't get a presigned POST URL for bucket '%s' "
                         "and object '%s'", bucket_name, object_key)
        raise
    return response


def generate_presigned_url(bucket_name, client_method, method_parameters,
                           expires_in):
    """
    Generate a presigned Amazon S3 URL that can be used to perform an action on
    a bucket. A presigned URL can be used for a limited time to let someone without
    an AWS account perform an action.

    Usage is shown in usage_demo at the end of this module.

    :param bucket_name: The name of the bucket on which the action can be performed.
    :param client_method: The name of the client method that the URL performs.
    :param method_parameters: The parameters of the specified client method.
    :param expires_in: The number of seconds the presigned URL is valid for.
    :return: The presigned URL.
    """
    s3 = get_s3()
    try:
        url = s3.meta.client.generate_presigned_url(
            ClientMethod=client_method,
            Params=method_parameters,
            ExpiresIn=expires_in
        )
        logger.info("Got presigned URL: %s", url)
    except ClientError:
        logger.exception("Couldn't get a presigned URL for bucket '%s' "
                         "and client method '%s'.",
                         bucket_name, client_method)
        raise
    return url


def usage_demo():
    """Demonstrates ways to use the functions in this module."""
    prefix = 'usage-demo-bucket-wrapper-'

    created_buckets = [create_bucket(prefix + str(uuid.uuid1()),
                                     s3_resource.meta.client.meta.region_name)
                       for _ in range(3)]
    for bucket in created_buckets:
        print(f"Created bucket {bucket.name}.")

    bucket_to_delete = created_buckets.pop()
    if bucket_exists(bucket_to_delete.name):
        print(f"Bucket exists: {bucket_to_delete.name}.")
    delete_bucket(bucket_to_delete)
    print(f"Deleted bucket {bucket_to_delete.name}.")
    if not bucket_exists(bucket_to_delete.name):
        print(f"Bucket no longer exists: {bucket_to_delete.name}.")

    buckets = [b for b in get_buckets() if b.name.startswith(prefix)]
    for bucket in buckets:
        print(f"Got bucket {bucket.name}.")

    bucket = created_buckets[0]
    grant_log_delivery_access(bucket.name)
    acl = get_acl(bucket.name)
    print(f"Bucket {bucket.name} has ACL grants: {acl.grants}.")

    put_rules = [{
        'AllowedOrigins': ['http://www.example.com'],
        'AllowedMethods': ['PUT', 'POST', 'DELETE'],
        'AllowedHeaders': ['*']
    }]
    put_cors(bucket.name, put_rules)
    get_rules = get_cors(bucket.name)
    print(f"Bucket {bucket.name} has CORS rules: {json.dumps(get_rules.cors_rules)}.")
    delete_cors(bucket.name)

    put_policy_desc = {
        'Version': '2012-10-17',
        'Id': str(uuid.uuid1()),
        'Statement': [{
            'Effect': 'Allow',
            'Principal': {'AWS': 'arn:aws:iam::111122223333:user/Martha'},
            'Action': [
                's3:GetObject',
                's3:ListBucket'
            ],
            'Resource': [
                f'arn:aws:s3:::{bucket.name}/*',
                f'arn:aws:s3:::{bucket.name}'
            ]
        }]
    }
    try:
        put_policy(bucket.name, put_policy_desc)
        policy = get_policy(bucket.name)
        print(f"Bucket {bucket.name} has policy {json.dumps(policy)}.")
        delete_policy(bucket.name)
    except ClientError as error:
        if error.response['Error']['Code'] == 'MalformedPolicy':
            print("Couldn't put the policy because the specified principal user does "
                  "not exist. For this request to succeed, you must replace the user "
                  "ARN with an actual AWS user.")
        else:
            raise

    put_rules = [{
        'ID': str(uuid.uuid1()),
        'Filter': {
            'And': {
                'Prefix': 'monsters/',
                'Tags': [{'Key': 'type', 'Value': 'zombie'}]
            }
        },
        'Status': 'Enabled',
        'Expiration': {'Days': 28}
    }]
    put_lifecycle_configuration(bucket.name, put_rules)
    get_rules = get_lifecycle_configuration(bucket.name)
    print(f"Bucket {bucket.name} has lifecycle configuration {json.dumps(get_rules)}.")
    delete_lifecycle_configuration(bucket.name)

    url = generate_presigned_url(
        bucket.name,
        'list_objects',
        {'Bucket': bucket.name},
        10
    )
    print(f"Generated a pre-signed URL that can be used to list the objects in "
          f"bucket {bucket.name}. The URL is {url}.")

    for bucket in created_buckets:
        bucket.delete()
        print(f"Deleted bucket {bucket.name}.")


def main():
    go = input("Running the usage demonstration uses your default AWS account "
               "credentials and might incur charges on your account. Do you want "
               "to continue (y/n)? ")
    if go.lower() == 'y':
        print("Starting the usage demo. Enjoy!")
        usage_demo()
    else:
        print("Thanks anyway!")


if __name__ == '__main__':
    main()



import sqlalchemy as sql
import sqlalchemy.orm as orm
import pandas as pd

def create_engine(conn_str,echo=False,**kwargs):
    engine=sql.create_engine(conn_str, echo=echo,**kwargs)
    return engine

def raw_sql(engine,query):
    with engine.connect() as con:
        rs = con.execute(query)
    return rs

def get_schemas(engine):
    inspector = sql.inspect(engine)
    return inspector.get_schema_names()

def get_metadata(engine,schema=None):
    meta = sql.MetaData()
    options={'oracle_resolve_synonyms':True} if engine.name=='oracle' else {}
    if schema is None and hasattr(engine,'conn_name') and engine.conn_name in schemas:
        schema=schemas[engine.conn_name]
    meta.reflect(bind=engine,schema=schema,**options)
    return meta

def get_session(engine):
    Session = orm.session.sessionmaker()
    return Session(bind=engine)

def get_table_names(meta):
    return list(meta.tables.keys())

def get_all_table_names(engine):
    res={}
    for schema in get_schemas(engine):
        meta=get_metadata(engine,schema=schema)
        if meta is not None:
            res[schema] = get_table_names(meta)
    return res

def query_to_df(q):
    try:
        df=pd.read_sql(q.statement, q.session.bind)
    except:
        df=pd.read_sql(str(q.statement), q.session.bind)
    return df

get_table_columns = lambda table: list(table.columns)
query_table=lambda table,session: session.query(table)
display_query=lambda q: display(HTML(query_to_df(q).to_html()))

def filter_for_rows(objs,filter_cols,objs_cols=None):
    if objs_cols is None:
        objs_cols=filter_cols
    filters=[]
    if orm.query.Query in type(objs).__mro__:
        for filter_col,objs_col in zip(filter_cols,objs_cols):
            filters.append(filter_col.in_(objs.with_entities(objs_col).subquery()))
    else:
        for filter_col,objs_col in zip(filter_cols,objs_cols):
            filters.append(filter_col.in_([getattr(obj,objs_col.name) for obj in objs]))
    return filters

def table_with_column(tables,col_name,assert_exists=False):
    tables=[table for table in tables if hasattr(table.columns,col_name)]
    table=tables[0] if len(tables)>0 else None
    if assert_exists:
        assert table is not None, f"No matching table in query exists with the column = {col_name}"
    return table

def convert_props_to_filters(tables,*args,exclude=False,**kwargs):
    args=list(args)
    for key,val in kwargs.items():
        props=key.split('__')
        table=table_with_column(tables,props[0])
        comparitors={'equal':lambda col,val:col==val,
                    'in':lambda col,val:col.in_(val)}
        if props[-1] in comparitors:
            comparison=comparitors[props[-1]]
            props=props[:-1]
        else:
            comparison=comparitors['equal']

        for i,prop in enumerate(props):
            col=getattr(table.columns,prop)
            lastprop=i==len(props)-1
            if not lastprop:
                fk=list(col.foreign_keys)[0]
                table=orm.aliased(fk.column.table)
                new_col=getattr(table.columns,fk.column.name)
                args.append(col==new_col)
            else:
                cur_arg=comparison(col,val)
                args.append(not cur_arg if exclude else cur_arg)
    return args



from .. import components
from ..components.base import *
from ..config import * #METRICS_PATH
import pandas as pd
import numpy as np
from ..utils import open_spectra, remove_baseline

class NewPage(components.Component):
    def __init__(self):
        self.layout=self.get_layout()
        self.set_callbacks()
    def get_layout(self):
        self.dropdown=components.Dropdown(multi=True,options=None,update_value_on_change=True,keep_old_values=True,value=None)
        self.uploader=components.Uploader(type='area',multi=True,allowed_extensions=SPECTRA_EXTENSIONS,upload_to=SPECTRA_PATH,dropdown=self.dropdown,keep_old_uploads=True)
        self.info=html.Div()

        return html.Div(['Example Div',self.uploader,self.dropdown,self.info])

    def set_callbacks(self):
        @self([self.info.children],[self.dropdown.value])
        def fcn(paths):
            msg=''
            for path in paths:
                spectra,wl=open_spectra(path)
                spectra_nb,background=remove_baseline(spectra)
                msg+=str(wl)
            return msg



from .components.base import *
from . import components

class Main(components.Component):
    def layout(self):
        version=os.environ.get('VERSION','unknown')
        self.overview=Overview()
        self.edit=Edit(overview=self.overview)
        self.amplitudes=Amplitudes(main_app=self)
        self.newpage=NewPage()
        self.tabs=components.Tabs({
                'Overview':self.overview,
                'Edit':self.edit,
                'Amplitudes':self.amplitudes,
                'NewPage':self.newpage
            })
        self.full_layout=html.Div([
            html.H1('ChipCHECK Verification'+(' - DEVELOPMENT BUILD' if os.environ.get('GIT_BRANCH','none').endswith('dev') else '')),
            self.tabs,
            html.Div(html.P(f'V{version}' if version !='unknown' else ''),style={'border-top':'1px solid rgba(0,0,0,.125)','width':'100%'},className='pt-3')
            ],className='pt-3')

        return self.full_layout



import inspect

def is_list(obj):
    return type(obj) in [list,tuple]

def copy_function_with_params(fcn, add_args=None, **kwargs):
    if add_args is None: add_args=[]
    if type(add_args) not in [list,tuple]: add_args=[add_args]
    new_args={key:val for key,val in kwargs.items() if key in fcn.__code__.co_varnames}
    varnames=add_args+[x for x in fcn.__code__.co_varnames if x not in new_args and x not in add_args]
    defaults={x:fcn.__defaults__[i] for i,x in enumerate(fcn.__code__.co_varnames[fcn.__code__.co_argcount-len(fcn.__defaults__):]) if x in varnames}
    fcn_args=add_args+[x for x in fcn.__code__.co_varnames[:fcn.__code__.co_argcount-len(fcn.__defaults__)] if x in varnames]
    def y():
        mylocals=locals()
        vars={x:mylocals.get(x) for x in varnames}
        vars.update(new_args)
        args=[vars[x] for x in fcn_args]
        kwargs={key:val for key,val in vars.items() if key not in fcn_args}
        return fcn(*args,**kwargs)

    #code(argcount, kwonlyargcount, nlocals, stacksize, flags, codestring,constants, names, varnames, filename, name, firstlineno,lnotab[, freevars[, cellvars]])
    fcn_code = types.CodeType(len(varnames),
                            y.__code__.co_kwonlyargcount,
                            y.__code__.co_nlocals,
                            y.__code__.co_stacksize,
                            y.__code__.co_flags,
                            y.__code__.co_code,
                            y.__code__.co_consts,
                            y.__code__.co_names,
                            tuple(varnames),
                            y.__code__.co_filename,
                            fcn.__code__.co_name,
                            y.__code__.co_firstlineno,
                            y.__code__.co_lnotab,
                            y.__code__.co_freevars,
                            y.__code__.co_cellvars)

    res=types.FunctionType(fcn_code, {**fcn.__globals__,**y.__globals__}, fcn.__code__.co_name,
                            argdefs=tuple([defaults[key] for key in varnames if key in defaults]),
                            closure=y.__closure__)
    res.__doc__=fcn.__doc__
    return res

class Spec(object):
    _defaults=None
    def __init__(self,fcn):
        self._fcn=fcn.__code__.co_varnames
        self.varnames=fcn.__code__.co_varnames
        self.name=fcn.__code__.co_name
        self.spec=inspect.getfullargspec(fcn)
        for spec_var,self_var in ({'args':'args','kwonlyargs':'kwargs','varargs':'varargs','varkw':'varkwargs','annotations':'annotations'}).items():
            setattr(self,self_var,list(spec_var) if is_list(spec_var) else spec_var)
        self.varnames_with_defaults=self.args[:(len(self.spec.defaults) if self.spec.defaults else 0)]+self.kwargs[:(len(self.spec.kwonlydefaults) if self.spec.kwonlydefaults else 0)]
        self.required_args=list(set(self.args+self.kwargs)-set(self.varnames_with_defaults))
    @property
    def defaults(self):
        if self._defaults is None:
            vals=(list(self.spec.defaults) if self.spec.defaults else [])+(list(self.spec.kwonlydefaults) if self.spec.kwonlydefaults else [])
            self._defaults={var:vals[self.varnames_with_defaults.index(var)] for var in self.varnames_with_defaults}
        return self._defaults
    def copy_function_with_params(self,add_args=None, **kwargs):
        return copy_function_with_params(self._fcn, add_args=add_args, **kwargs)
    def need_more_inputs(self,available_inputs=None,return_missing=False):
        if available_inputs is None: available_inputs=['self']
        missing=list(set(self.required_args)-set(available_inputs))
        return missing if return_missing else len(missing)==0

def need_more_inputs(fcn,available_inputs=None,return_missing=False):
    return Spec(fcn).need_more_inputs(available_inputs=available_inputs,return_missing=return_missing)

def describe_fcn(fcn):
    ''' returns a spec for the function with the properties: name, varnames, defaults (for both args and kwargs), args, kwargs, vararg, varkwarg, variables={"name":{"default","type"}} (type is one of [arg, kwarg, vararg, varkwarg])
    '''
    Spec(fcn)



import os
import subprocess
import sys
import pkg_resources
import requests
from datetime import datetime

from .utils import get_cwd

class Lazy_Module(object):
    _released=None
    def __init__(self,module,name=None):
        self._mod_module=module if isinstance(module,Module) else Module(module)
        self._mod_name=name if name else self._mod_module.name
        self.self_names=['_mod_module','_mod_name','_released','_release']
    @property
    def _release(self):
        if not self._released:
            module=self._mod_module.load()
            globals()[self._mod_name]=module
            self._released=module
        return self._released
    def __getattribute__(self,name):
        try:
            return object.__getattribute__(self,name)
        except:
            released=object.__getattribute__(self,'_release')
            return getattr(released,name)
    def __setattribute__(self,name,val):
        try:
            return object.__setattribute__(self,name,val)
        except:
            released=object.__getattribute__(self,'_release')
            return setattr(released,name,val)
    def __call__(self,*args,**kwargs):
        return self._release(*args,**kwargs)
    def __getitem__(self,val):
        return self._release[val]

def lazy_import(obj,*objs,from_path=None):
    objs=[obj]+list(objs)
    if from_path:
        objs=[os.path.join(from_path,x) for x in objs]
    mods=[Lazy_Module(x) for x in objs]
    return mods if len(mods)>1 else mods[0]

class Version(object):
    major=minor=micro=None
    def __init__(self,val):
        labels={0:'major',1:'minor',2:'micro'}
        self.version=val
        if isinstance(val,str):
            self.is_dist=True
            vals=val.split('.')
            for i in range(len(vals)):
                setattr(self,labels[i],int(vals[i]))
        elif isinstance(val,float):
            self.is_dist=False
            d=datetime.fromtimestamp(val)
            for i,attr in enumerate(['year','month','day']):
                setattr(self,labels[i],getattr(d,attr))
    def __repr__(self):
        return '.'.join([str(getattr(self,x)) for x in ['major','minor','micro'] if getattr(self,x) is not None])
    def __str__(self):
        return f'Version ({self.__repr__()})'
    def __lt__(self,val):
        return self.version<val
    def __gt__(self,val):
        return self.version>val
    def __eq__(self,val):
        return self.version==val

class Module(object):
    _path=None
    _name=None
    _module=None
    _dist=None
    def __init__(self,package):
        self.package=package
    def _set_name_path(self):
        if os.path.isdir(self.package) and os.path.isfile(os.path.join(self.package,'__init__.py')):
            self._path=self.package
        elif self.package.endswith('__init__.py') and os.path.isfile(self.package):
            self._path=os.path.dirname(self.package)
        else:
            self._path,self._name=self.find_module(self.package)
        if self._path and not self._name:
            self._name=os.path.basename(self._path)
    @property
    def path(self):
        if not self._path: self._set_name_path()
        return self._path
    @property
    def name(self):
        if not self._name: self._set_name_path()
        return self._name
    @property
    def mtime(self):
        return os.path.getmtime(self.path)
    @property
    def file_loader(self):
        return self._get_file_loader(os.path.dirname(self.path),self.name)
    def _get_file_loader(self,path,name):
        if os.path.isdir(path):
            finder=pkg_resources.get_importer(path)
            return finder.find_module(name)
    @property
    def is_dist(self):
        return self.dist is not None
    @property
    def dist(self):
        if not self._dist:
            try:
                cur_dist = pkg_resources.get_distribution(self.name)
                if os.path.normpath(self.path)==os.path.normpath(cur_dist.location):
                    self._dist = cur_dist
            except:
                pass
        return self._dist
    def find_module(self,name):
        for path in list(sys.path):
            loader=self._get_file_loader(path,name)
            if loader:
                return os.path.dirname(loader.path),loader.name
    @property
    def pypi_info(self):
        return requests.get(f'https://pypi.org/pypi/{self.name}/json').json()['info']
    @property
    def version(self):
        return Version(self.dist.version if self.is_dist else self.mtime)
    @property
    def is_latest(self):
        return self.version==self.latest_version
    @property
    def latest_version(self):
        if self.is_dist:
            return Version(self.pypi_info['version'])
    def load(self):
        if not self._module:
            try:
                self._module=self.file_loader.load_module()
            except Exception as e:
                print(e)
                self._module=__import__(self.name)
        return self._module
    def import_module(self):
        return self.load()
    @property
    def children(self):
        return Modules(paths=[self.path])

class Modules(object):
    def __init__(self,paths=None):
        if paths is None: paths=list(sys.path)
        self.paths=paths
        self._modules=None
    @property
    def modules(self):
        return list(self.get_modules().keys())
    def get_modules(self,refresh=False):
        if self._modules is None or refresh:
            self._modules={}
            for root in self.paths:
                if root=='':
                    root=get_cwd()
                if os.path.isdir(root):
                    for obj in os.listdir(root):
                        path=os.path.join(root,obj)
                        if os.path.isdir(path) and os.path.isfile(os.path.join(path,'__init__.py')):
                            module=Module(path)
                            self._modules[module.name]=module
        return self._modules
    def get(self,name):
        modules=self.get_modules()
        return modules[name] if name in modules else None
    def __getitem__(self,name):
        return self.get(name)
    def __getattribute__(self,name):
        try:
            return object.__getattribute__(self,name)
        except:
            return self.get(name)

def install(package):
    return subprocess.check_call([sys.executable, "-m", "pip", "install", "--upgrade", package])



import dash
from dash.dependencies import Input, Output, State, MATCH, ALL
import uuid

from dash.development.base_component import Component as BaseComponent, ComponentMeta
from plotly.basedatatypes import BaseFigure

import inspect
import threading
from copy import copy

BASE_COMPONENTS=[BaseComponent,ComponentMeta]

def generate_unique_id():
    return uuid.uuid4().hex[:6].upper()

def update_class(obj,new_class):
    for key,val in dict(new_class.__dict__).items():
        if key not in ['__module__','__dict__','__weakref__','__doc__']:
            setattr(obj,key,val)

to_list = lambda obj: [] if obj is None else list(obj) if type(obj) in [list,tuple] and (len(obj)==0 or type(obj[0]) in [list,tuple]) else [obj]

def iterable(obj):
    try:
        _=iter(obj)
        return True
    except:
        return False

class Callback(object):
    contexts = threading.local()
    fcn=outputs=inputs=states=format_output=None
    def __init__(self,fcn=None,outputs=None,inputs=None,states=None,format_output=None):
        self.fcn=fcn
        self.outputs=to_list(outputs)
        self.inputs=to_list(inputs)
        self.states=to_list(states)
        self.format_output=format_output
    def __str__(self):
        return f'Callback(fcn={self.fcn}, outputs={self.outputs}, inputs={self.inputs}, states={self.states}, format_output={self.format_output})'
    def __repr__(self):
        return str(self)
    def register(self):
        type(self).get_callbacks().append(self)
    def is_valid(self):
        valid_in_out=lambda vals: (isinstance(vals,list) and len(vals)>0 and
                all([type(x) in [list,tuple] and len(x)==2 for x in vals]))
        checks={'fcn':lambda x: (x is not None and callable(x)),
            'outputs':valid_in_out,
            'inputs':valid_in_out,
            'states':lambda vals: (isinstance(vals,list)),
            'format_output':lambda x: (x is None or callable(x))}
        return all([fcn(getattr(self,prop)) for prop,fcn in self.checks.items()])
    @classmethod
    def get_callbacks(cls):
        if not hasattr(cls.contexts, 'callbacks'):
            cls.contexts.callbacks = []
        return cls.contexts.callbacks

def Call(outputs, inputs, states=None, format_output=None):
    def rtn(fcn):
        Callback(fcn=fcn,outputs=outputs,inputs=inputs,states=states,format_output=format_output).register()
    return rtn

class Item(object):
    dynamic_id=obj=indexes=None
    def __init__(self,parent,name,obj):
        objID=f'{parent.id}-{name}'
        if inspect.isclass(obj):
            self.dynamic_id=objID
        else:
            obj.id=objID
        self.obj=obj
        self.indexes=[]
    def __call__(self,*args,index=None,**kwargs):
        assert self.dynamic_id is not None, 'Can only create dynamic components passed as an object class'
        if index is None:
            index=str(len(self.indexes))
        obj=self.obj(*args,**kwargs)
        obj.id={'type':self.dynamic_id,'index':index}
        self.indexes.append(index)
        return obj
    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except Exception as e:
            if self.dynamic_id is not None:
                prop_type=ALL
                if name.endswith('_match'):
                    name=name.replace('_match','')
                    prop_type=MATCH
                elif name.endswith('_all'):
                    name=name.replace('_all','')
                return ({'type':self.dynamic_id,'index':prop_type},name)
            elif name in self.obj._prop_names:
                return (self.obj.id,name)
            else:
                raise e
    def __setattr__(self, name, val):
        obj=self.obj if hasattr(self.obj,name) else self
        object.__setattr__(obj,name, val)

class Component(object):
    properties=callbacks=_id=None
    def class_instances(self):
        instances=type(self).get_instances()
        if self.__class__.__name__ not in instances:
            instances[self.__class__.__name__]=[]
        if self not in instances[self.__class__.__name__]:
            instances[self.__class__.__name__].append(self)
        return instances[self.__class__.__name__]
    @property
    def id(self):
        if not self._id:
            instances=self.class_instances()
            self._id=f'{self.__class__.__name__}-{len(instances)}'
        return self._id
    def __call__(self, outputs, inputs, states=None, format_output=None):
        def rtn(fcn):
            self.add_callback(fcn=fcn,outputs=outputs,inputs=inputs,states=states,format_output=format_output)
        return rtn
    def className(self,names):
        class_names=self.layout.obj.className
        if class_names is None:
            class_names=""
        elif class_names!="" and not class_names.endswith(' '):
            class_names+=' '
        self.layout.obj.className=class_names+names.strip()
        return self
    def add_callback(self,fcn, outputs, inputs, states=None, format_output=None):
        if not self.callbacks: self.callbacks=[]
        self.callbacks.append(Callback(fcn=fcn,outputs=outputs,inputs=inputs,states=states,format_output=format_output))
        self.callbacks[-1].register()
    def __getattribute__(self, name):
        properties=object.__getattribute__(self, 'properties')
        if isinstance(properties,dict) and name in properties:
            return properties[name]
        return object.__getattribute__(self, name)
    def __setattr__(self, name, val):
        if any([isinstance(val,x) for x in BASE_COMPONENTS]):
            object.__setattr__(self, name, Item(self,name,val))
        else:
            return object.__setattr__(self, name, val)
    def process_layout_children(self,layout):
        def get_layout(val):
            #print('3)',val)
            if val is not None:
                if Component in type(val).__mro__:
                    val=val.generate_layout()
                elif isinstance(val,Item):
                    val=self.generate_layout(val.obj)
                else:
                    val=self.generate_layout(val)
            return val
        #print('1)',layout)
        if hasattr(layout,'children'):
            children=layout.children
            #print('2)',type(children),children)
            if children is not None:
                if isinstance(children,list):
                    layout.children=[get_layout(x) for x in children]
                else:
                    layout.children=get_layout(children)
    def generate_layout(self,layout=None):
        if layout is None:
            if isinstance(self.layout,Item):
                layout=self.layout.obj
            elif callable(self.layout):
                layout=self.layout()
                if isinstance(layout,Item):
                    layout=layout.obj
            else:
                layout=self.layout

        self.process_layout_children(layout)

        return layout
    @classmethod
    def get_instances(cls):
        if not hasattr(cls, 'instances'):
            cls.instances={}
        return cls.instances


#for comp in BASE_COMPONENTS:
#    update_class(comp,Component)
def inputs_match_trigger(inputs,states,callback_trigger):
    inputs=to_list(inputs)
    states=to_list(states)
    trigger_ids=[x['prop_id'] for x in callback_trigger] #['{"index":"Transfer","type":"folder-item-B4DAD1"}.n_clicks']
    input_in_trigger = lambda x,trigger: ((x[0]['type'] if isinstance(x[0],dict) and 'type' in x[0] else str(x[0])) in trigger) and (x[1] in trigger)
    return any([any([input_in_trigger(x,trigger_x) for trigger_x in trigger_ids]) for x in inputs+states]) #[{'prop_id': 'alert-interval-88C021.n_intervals', 'value': 5}]

def generate_callback(app,outputs,inputs,states,callbacks):
    @app.callback([Output(x[0],x[1]) for x in outputs],
                [Input(x[0],x[1]) for x in inputs],
                [State(x[0],x[1]) for x in states])
    def rtn(*values):
        values=list(values)
        inputs_states=inputs+states
        res=values[-len(outputs):]
        for callback in callbacks:
            if inputs_match_trigger(callback.inputs,callback.states,dash.callback_context.triggered):
                values_pos=[inputs_states.index(x) for x in (callback.inputs+callback.states)]
                output_pos=[outputs.index(x) for x in (callback.outputs)]
                fcn_res=callback.fcn(*[values[i] for i in values_pos])
                if fcn_res is not None:
                    callback_outputs=list(tuple(fcn_res)) if type(fcn_res) in [list,tuple] else [fcn_res]
                    if callback.format_output:
                        callback_outputs=callback.format_output(callback_outputs)
                    for callback_i,output_i in enumerate(output_pos):
                        res[output_i]=callback_outputs[callback_i]
                    break
        return tuple(res)

def generate_all_callbacks(app,verbose=0):
    callbacks=Callback.get_callbacks()
    if len(callbacks)==0 and callable(app.layout):
        _=app.layout()
        callbacks=Callback.get_callbacks()
    combined_callbacks=[] #list of {'outputs':[],'callbacks':[]}

    for callback in callbacks:
        added_to=None
        for i,combined_callback in enumerate(combined_callbacks):
            if combined_callback is not None and any([x in combined_callback['outputs'] for x in callback.outputs]):
                if added_to is not None:
                    combined_callbacks[added_to]['outputs']+=[x for x in combined_callbacks[i]['outputs'] if x not in combined_callbacks[added_to]['outputs']]
                    combined_callbacks[added_to]['callbacks'].extend(combined_callbacks[i]['callbacks'])
                    combined_callbacks[i]=None
                else:
                    added_to=i
                    combined_callbacks[i]['outputs']+=[x for x in callback.outputs if x not in combined_callbacks[i]['outputs']]
                    combined_callbacks[i]['callbacks'].append(callback)
        if added_to is None:
            combined_callbacks.append({'outputs':copy(callback.outputs),'callbacks':[callback]})

    for combined_callback in [x for x in combined_callbacks if x is not None]:
        output_states=combined_callback['outputs']

        unique_inputs=[];unique_states=[]
        for callback in combined_callback['callbacks']:
            unique_inputs.extend([x for x in callback.inputs if x not in unique_inputs])
            unique_states.extend([x for x in callback.states if x not in (unique_states+output_states)])
        unique_states.extend(output_states)
        unique_inputs_states=unique_inputs+unique_states

        cur_outputs=copy(output_states);cur_inputs=copy(unique_inputs);cur_states=copy(unique_states);cur_callbacks=copy(combined_callback["callbacks"])
        if verbose>0:
            app.logger.warning(f'All Outputs = {cur_outputs}\nAll Inputs = {cur_inputs}\nAll States = {cur_states}')
            app.logger.warning("Callbacks = \n\t"+"\n\t".join([str(x) for x in cur_callbacks])+"\n\n")

        generate_callback(app,cur_outputs,cur_inputs,cur_states,callbacks=cur_callbacks)



import os
from io import BytesIO
import gridfs
import pymongo
from pymongo import InsertOne, DeleteMany, ReplaceOne, UpdateOne, UpdateMany, IndexModel
from .process import get_buffer_metas, df_to_buffer, buffer_to_df
from .utils import get_client, to_list, obj_to_buffer, buffer_to_obj

INDEX_OPTIONS={x:getattr(pymongo,x) for x in ['ASCENDING', 'DESCENDING', 'GEO2D', 'GEOHAYSTACK', 'GEOSPHERE', 'HASHED', 'TEXT']}

class COLLECTION(object):
    _properties=None
    def __init__(self,collections,id):
        self._collections=collections
        self._id=id
    @property
    def properties(self):
        if self._properties is None:
            self._properties=self._collections._db_collection.find_one({'_id':self._id})
        return self._properties
    def __repr__(self):
        return f'{self._collections.name.capitalize()}({self._id})'
    def __str__(self):
        return str(self.properties)
    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except:
            assert name in self.properties, f'{self.__repr__()} does not have a meta attribute "{name}"'
            return self.properties[name]
    def update(self,*args,**kwargs):
        updates=self._collections._get_updates(*args,**kwargs)
        return self._collections._db_collection.update_one({'_id':self._id},updates)

    def delete(self,properties=None):
        if properties is not None:
            if type(properties) not in [list,tuple]: properties=[properties]
            unsets={}
            for prop in properties:
                unsets[prop]=1
                if self._properties is not None and prop in self._properties:
                    del self.properties[prop]
            return self._collections._db_collection.update_one({'_id':self._id},{'$unset':unsets})
        else:
            _id=self._id
            res = self._collections._db_collection.delete_one({'_id':_id})
            if hasattr(self._collections,'_gridfs'):
                res2=self._collections._gridfs._GridFS__chunks.delete_one({'files_id':_id})
                res=(res,res2)
            del self
            return res
    def __delete__(self):
        return self.delete()

class COLLECTIONS(object):
    def __init__(self,collection,required_meta=None,create_index=None):
        self._db_collection=collection
        self._required_meta=required_meta
        self.child_class=COLLECTION
        if create_index is not None:
            options={}
            if isinstance(create_index,str):
                indexes=create_index
            elif isinstance(create_index,dict):
                indexes=[]
                for field,props in create_index.items():
                    indexes.append((field,INDEX_OPTIONS[props['type']]) if 'type' in props else field)
                    if 'options' in props:
                        options.update(props['options'])
            else:
                raise Exception('create_index must be one of [str,dict]')
            self._db_collection.create_index(indexes,**options)
    def __getattr__(self,name):
        try:
            object.__getattr__(self,name)
        except:
            return getattr(self._db_collection,name)

    @property
    def name(self):
        return self._db_collection._Collection__name
    def get(self,*args,**query):
        query.update(dict(pair for d in args if isinstance(d,dict) for pair in d.items()))
        return [self.child_class(self,x['_id']) for x in self._db_collection.find(query,projection='_id')]
    def _check_meta(self,query=None,*args):
        if self._required_meta is None:
            return True
        if query is None:
            query={}
        else:
            query=query.copy()
        for arg in args:
            query.update(arg['$set'] if '$set' in arg else arg)
        return all([x in query for x in self._required_meta])
    def _get_updates(self,*args,**kwargs):
        kwargs.update(dict(pair for d in args if isinstance(d,dict) for pair in d.items()))
        updates={}
        for key,val in kwargs.items():
            if not key.startswith('$'):
                if '$set' not in updates:
                    updates['$set']={}
                updates['$set'][key]=val
            else:
                updates[key]=val
        return updates
    def update_or_create(self,query=None,*args,**kwargs):
        if query is None: query={}
        updates=self._get_updates(*args,**kwargs)
        if len(query)==0 and '$set' in updates:
            query=updates['$set']
        assert len(query)>0, "No attributes provided for the record"
        if len(updates)==0: updates['$set']=query

        if not isinstance(query,list):
            query=[query]

        ops=[];combined_query={}
        for cur_query in query:
            if self._check_meta(cur_query,updates):
                ops.append(UpdateMany(cur_query, updates,upsert=True))
                for key,val in cur_query.items():
                    if key in combined_query:
                        if combined_query[key]!=val:
                            if not isinstance(combined_query[key],dict) or '$or' not in combined_query[key]:
                                combined_query[key]={'$or':[combined_query[key]]}
                            if val not in combined_query[key]['$or']:
                                combined_query[key]['$or'].append(val)
                    else:
                        combined_query[key]=val

        #bulk write ops
        ids=[]
        if len(ops)>0:
            result=self._db_collection.bulk_write(ops)
            ids=[x['_id'] for x in self._db_collection.find(combined_query,projection='_id')]

        return [self.child_class(self,x) for x in ids]

class FILE(COLLECTION):
    _buffer=None
    _file=None
    @property
    def buffer(self):
        if self._buffer is None:
            self._buffer= self._collections._gridfs.find_one({'_id':self._id})
        return self._buffer
    @property
    def file(self):
        if self._file is None:
            buffer=self.buffer
            props=self.properties
            varnames=self._collections._buffer_to_file.__code__.co_varnames
            send_props={key:val for key,val in props.items() if 'kwargs' in varnames or key in varnames}
            if 'props' in varnames: send_props['props']=props
            if 'properties' in varnames: send_props['properties']=props
            self._file=self._collections._buffer_to_file(buffer,**send_props)
        return self._file
    @property
    def obj(self):
        return self.file
    def save_to(self,path):
        if os.path.isdir(path):
            filename=self.info.filename
            if filename is None:
                filename=f'{self._collections.name.capitalize()}_{self._id}'
            path=os.path.join(path,filename)
        with open(path,'wb') as f:
            f.write(self.buffer.read())
        return path

class FILES(COLLECTIONS):
    def __init__(self,gridfs,extensions=None,path_processor=None,buffer_to_file=None,file_to_buffer=None,**kwargs):
        super().__init__(gridfs._GridFS__files,**kwargs)
        self._gridfs=gridfs
        self._buffer_to_file=buffer_to_obj if buffer_to_file is None else buffer_to_file
        self._extensions=extensions
        self._path_processor=path_processor
        self._file_to_buffer=obj_to_buffer if file_to_buffer is None else file_to_buffer
        self.child_class=FILE

    def update_or_create(self,files=None,query=None,*args,**kwargs):
        ids=[]
        if files is not None:
            if query is None: query={}
            updates={}
            for key,val in kwargs.items():
                if key=='$set':
                    updates.update(val)
                elif not key.startswith('$'):
                    updates[key]=val
            updates.update(dict(pair for d in args if isinstance(d,dict) for pair in d.items()))
            if len(query)==0:
                query=updates
                updates={}

            if not isinstance(files,list):
                files=[files]

            buffers=[]
            for file in files:
                if isinstance(file,str) and os.path.exists(file):
                    buffers.extend(self.process_path(file))
                else:
                    buffers.append(self._file_to_buffer(file))

            for buffer,meta in get_buffer_metas(buffers):

                #query and meta
                cur_query={**query,**meta}
                cur_meta={**cur_query,**updates}

                #Ensure we have required_meta
                if self._check_meta(cur_meta):
                    #delete file if it exists
                    if len(cur_query)>0:
                        existing=self._gridfs.find_one(cur_query)
                        if existing is not None:
                            self._gridfs.delete(existing._id)

                    #Store buffer and close
                    ids.append(self._gridfs.put(buffer,**cur_meta))
                    BytesIO.close(buffer)
            return [self.child_class(self,x) for x in ids]
        else:
            return super().update_or_create(query=query,*args,**kwargs)

    def process_file(self,path):
        if self._extensions is not None and not any([path.lower().endswith(x) for x in self._extensions]):
            return None
        return obj_to_buffer(path)

    def process_path(self,path):
        buffers=[]
        if self._path_processor:
            files=self._path_processor(path)
            buffers=[self._file_to_buffer(file) for file in files]
        else:
            paths=[path] if os.path.isfile(path) else [os.path.join(path,file) for file in os.listdir(path)]
            for cur_path in paths:
                buffer=self.process_file(cur_path)
                if buffer is not None:
                    buffers.append(buffer)
        return buffers

class DB(object):
    _collections={}
    def __init__(self,DB_name,url='l3-37:27017/',collections=None,username=None,password=None,srv=False):
        client = get_client(url=url,username=username,password=password,srv=srv)
        self.DB_name=DB_name
        self._db=client[DB_name]
        if collections is not None:
            #collections must be a dict of collections with relavent properties
            self._initialize_collections(collections)
    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except:
            if name not in self._collections:
                self._collections[name]=COLLECTIONS(getattr(self._db,name))
            return self._collections[name]
    def GridFS(self,name,collection_name=None,file_type=None,**kwargs):
        if name not in self._collections:
            if file_type is not None:
                if file_type.lower()=='dataframe':
                    kwargs['buffer_to_file']=buffer_to_df
                    kwargs['file_to_buffer']=df_to_buffer
                    kwargs['extensions']=['.csv']
                elif file_type.lower()=='image':
                    kwargs['extensions']=['.jpg','png','.jpeg']
            if collection_name is None:
                collection_name=name
            self._collections[name]=FILES(gridfs.GridFS(self._db,collection=collection_name),**kwargs)
        return self._collections[name]
    def Collection(self,name,collection_name=None,**kwargs):
        if name not in self._collections:
            if collection_name is None:
                collection_name=name
            self._collections[name]=COLLECTIONS(getattr(self._db,collection_name),**kwargs)
        return self._collections[name]
    def _initialize_collections(self,descriptions):
        common_props=['collection_name','required_meta','create_index']
        FILES_props=['file_type','extensions','path_processor','buffer_to_file','file_to_buffer']
        for name,props in descriptions.items():
            collection_type=props['collection_type'].lower() if 'collection_type' in props else 'standard'
            if any([x in props for x in FILES_props]):
                collection_type='gridfs'

            #get common props:
            kwargs={x:(props[x] if x in props else None) for x in common_props}

            if collection_type=='standard':
                _=self.Collection(name,**kwargs)
            elif collection_type=='gridfs':
                kwargs.update({x:(props[x] if x in props else None) for x in FILES_props})
                _=self.GridFS(name,**kwargs)
    def __repr__(self):
        return f'DB({self.DB_name})'
    def __str__(self):
        return f'{self.DB_name} with {self.tests._db_collection.count_documents({})} tests, {self.tables._db_collection.count_documents({})} tables, {self.images._db_collection.count_documents({})} images, and {self.files._db_collection.count_documents({})} files'



from . import db, classes

get_oracle_conn = lambda user,password,host,port,server: f'oracle+cx_oracle://{user}:{password}@{host}:{port}/{server}'
oracle_connections={'buildware':dict(user='gpms_read',password='readonly',host='ackbar',port='1521',server='prod112'),
                    'halifax':dict(user='gpms_read',password='readonly',host='tie',port='1521',server='shwtr102')}
oracle_schemas={
                'buildware':'gpms_owner',
                'halifax':'gpms_owner',
                }




class Analyzer(classes.DB):
    def __init__(self,db_params=None,schema=None):
        if isinstance(db_params,str):
            assert db_params in oracle_connections, f"{db_params} is not available in standard definitions, please provide the user, password, host, port and server instead."
            if schema is None and db_params in oracle_schemas:
                schema=oracle_schemas[db_params]
            db_params=oracle_connections[db_params]
        assert isinstance(db_params,dict), f"A dictionary of the databe connection is required. Instead recieved db as {type(db_params)}"
        conn_params=['user', 'password', 'host', 'port', 'server']
        missing_params=[x for x in conn_params if x not in db_params]
        assert len(missing_params)==0, f"Missing connection parameters: {', '.join(missing_params)}"
        self.db_params={key:val for key,val in db_params.items() if key in conn_params}
        conn_str=get_oracle_conn(**self.db_params)
        if schema is None:
            schema=self.db_params['user']
        super().__init__(conn_str,schema=schema)



import os

env_file_variables=None
def get_variable(name,default=None):
    global env_file_variables
    if env_file_variables is None:
        env_file_variables={}
        for file in os.listdir('.'):
            if file.lower().endswith('.env'):
                with open(file,'r') as f:
                    lines=f.readlines()
                env_file_variables={splits[0].strip():"".join([x.strip() for x in splits[1:]]) for line in lines if '=' in line for splits in line.split('=')}
                break

    if name in env_file_variables: return env_file_variables[name]
    return os.environ.get(name,default)

class Config(object):
    SECRET_KEY=get_variable('SECRET_KEY','adsfyiu3S3!jE%$axbjhwa195sxc@S')
    def get(self,name,default=None):
        if hasattr(self,name):
            return getattr(self,name)
        return default
